@@ -52,7 +52,7 @@ module LLVM.Core.Instructions(
     bitcast,
     -- * Comparison
     CmpPredicate(..), IntPredicate(..), FPPredicate(..),
-    CmpRet, CmpResult, CmpValueResult,
+    CmpRet, CmpResult,
     cmp, pcmp, icmp, fcmp,
     select,
     -- * Fast math
@@ -81,7 +81,7 @@ import qualified LLVM.Core.Util as U
 import qualified LLVM.Core.Proxy as LP
 import qualified LLVM.Core.CodeGen as CodeGen
 import LLVM.Core.Instructions.Private
-            (ValueCons, unValue, convert,
+            (ValueCons, unValue, convert, convertValue,
              unop, unopValue, binopValue, proxyFromValuePtr,
              FFIBinOp, FFIConstBinOp,
              GetField, FieldType, GetElementPtr, ElementPtrType,
@@ -103,6 +103,7 @@ import LLVM.FFI.Core (IntPredicate(..), FPPredicate(..))
 import qualified Type.Data.Num.Decimal.Number as Dec
 import Type.Data.Num.Decimal.Literal (d1)
 import Type.Data.Num.Decimal.Number ((:<:), (:>:))
+import Type.Data.Bool (False)
 import Type.Base.Proxy (Proxy)
 
 import qualified Foreign
@@ -401,22 +402,26 @@ mul =
       IntegerType  -> binopValue FFI.buildMul
       FloatingType -> binopValue FFI.buildFMul
 
-iadd, isub, imul ::
+iadd, isub ::
     (ValueCons2 value0 value1, IsInteger a) =>
     value0 a -> value1 a -> CodeGenFunction r (BinOpValue value0 value1 a)
 iadd = binop FFI.constAdd FFI.buildAdd
 isub = binop FFI.constSub FFI.buildSub
-imul = binop FFI.constMul FFI.buildMul
 
-iaddNoWrap, isubNoWrap, imulNoWrap ::
+imul :: (IsInteger a) => Value a -> Value a -> CodeGenFunction r (Value a)
+imul = binopValue FFI.buildMul
+
+iaddNoWrap, isubNoWrap ::
     (ValueCons2 value0 value1, IsInteger a) =>
     value0 a -> value1 a -> CodeGenFunction r (BinOpValue value0 value1 a)
 iaddNoWrap =
     sbinop FFI.constNSWAdd FFI.buildNSWAdd FFI.constNUWAdd FFI.buildNUWAdd
 isubNoWrap =
     sbinop FFI.constNSWSub FFI.buildNSWSub FFI.constNUWSub FFI.buildNUWSub
+
+imulNoWrap :: (IsInteger a) => Value a -> Value a -> CodeGenFunction r (Value a)
 imulNoWrap =
-    sbinop FFI.constNSWMul FFI.buildNSWMul FFI.constNUWMul FFI.buildNUWMul
+    sbinopValue FFI.buildNSWMul FFI.buildNUWMul
 
 -- | signed or unsigned integer division depending on the type
 idiv :: (IsInteger a) => Value a -> Value a -> CodeGenFunction r (Value a)
@@ -449,20 +454,21 @@ fdiv = binopValue FFI.buildFDiv
 frem :: (IsFloating a) => Value a -> Value a -> CodeGenFunction r (Value a)
 frem = binopValue FFI.buildFRem
 
-shl, lshr, ashr, and, or, xor ::
+xor ::
     (ValueCons2 value0 value1, IsInteger a) =>
     value0 a -> value1 a -> CodeGenFunction r (BinOpValue value0 value1 a)
-shl  = binop FFI.constShl  FFI.buildShl
-lshr = binop FFI.constLShr FFI.buildLShr
-ashr = binop FFI.constAShr FFI.buildAShr
-and  = binop FFI.constAnd  FFI.buildAnd
-or   = binop FFI.constOr   FFI.buildOr
 xor  = binop FFI.constXor  FFI.buildXor
 
+shl, lshr, ashr, and, or :: Value a -> Value a -> CodeGenFunction r (Value a)
+shl  = binopValue FFI.buildShl
+lshr = binopValue FFI.buildLShr
+ashr = binopValue FFI.buildAShr
+and  = binopValue FFI.buildAnd
+or   = binopValue FFI.buildOr
+
 shr ::
-    (ValueCons2 value0 value1, IsInteger a) =>
-    value0 a -> value1 a -> CodeGenFunction r (BinOpValue value0 value1 a)
-shr = sbinop FFI.constAShr FFI.buildAShr FFI.constLShr FFI.buildLShr
+    (IsInteger a) => Value a -> Value a -> CodeGenFunction r (Value a)
+shr = sbinopValue FFI.buildAShr FFI.buildLShr
 
 sbinopValue ::
     forall a b r.
@@ -508,12 +514,10 @@ ineg = unop FFI.constNeg FFI.buildNeg
 
 inegNoWrap ::
     forall value a r.
-    (ValueCons value, IsInteger a) =>
+    (ValueCons value, IsInteger a, Signed a ~ False) =>
     value a -> CodeGenFunction r (value a)
 inegNoWrap =
-   if isSigned (LP.Proxy :: LP.Proxy a)
-     then unop FFI.constNSWNeg FFI.buildNSWNeg
-     else unop FFI.constNUWNeg FFI.buildNUWNeg
+   unop FFI.constNSWNeg FFI.buildNSWNeg
 
 fneg ::
     (IsFloating a) =>
@@ -621,107 +625,138 @@ trunc = convert FFI.constTrunc FFI.buildTrunc
 
 -- | Zero extend a value to a wider width.
 -- If possible, use 'ext' that chooses the right padding according to the types
-zext :: (ValueCons value, IsInteger a, IsInteger b, ShapeOf a ~ ShapeOf b, IsSized a, IsSized b, SizeOf a :<: SizeOf b)
-     => value a -> CodeGenFunction r (value b)
-zext = convert FFI.constZExt FFI.buildZExt
+zext ::
+    (IsInteger a, IsInteger b, ShapeOf a ~ ShapeOf b,
+     IsSized a, IsSized b, SizeOf a :<: SizeOf b) =>
+    Value a -> CodeGenFunction r (Value b)
+zext = convertValue LP.Proxy FFI.buildZExt
 
 -- | Sign extend a value to wider width.
 -- If possible, use 'ext' that chooses the right padding according to the types
-sext :: (ValueCons value, IsInteger a, IsInteger b, ShapeOf a ~ ShapeOf b, IsSized a, IsSized b, SizeOf a :<: SizeOf b)
-     => value a -> CodeGenFunction r (value b)
-sext = convert FFI.constSExt FFI.buildSExt
+sext ::
+    (IsInteger a, IsInteger b, ShapeOf a ~ ShapeOf b,
+     IsSized a, IsSized b, SizeOf a :<: SizeOf b) =>
+    Value a -> CodeGenFunction r (Value b)
+sext = convertValue LP.Proxy FFI.buildSExt
 
 -- | Extend a value to wider width.
 -- If the target type is signed, then preserve the sign,
 -- If the target type is unsigned, then extended by zeros.
-ext :: forall value a b r. (ValueCons value, IsInteger a, IsInteger b, ShapeOf a ~ ShapeOf b, Signed a ~ Signed b, IsSized a, IsSized b, SizeOf a :<: SizeOf b)
-     => value a -> CodeGenFunction r (value b)
+ext ::
+    forall a b r.
+    (IsInteger a, IsInteger b, ShapeOf a ~ ShapeOf b, Signed a ~ Signed b,
+     IsSized a, IsSized b, SizeOf a :<: SizeOf b) =>
+    Value a -> CodeGenFunction r (Value b)
 ext =
    if isSigned (LP.Proxy :: LP.Proxy b)
-     then convert FFI.constSExt FFI.buildSExt
-     else convert FFI.constZExt FFI.buildZExt
+     then convertValue LP.Proxy FFI.buildSExt
+     else convertValue LP.Proxy FFI.buildZExt
 
 -- | It is 'zext', 'trunc' or nop depending on the relation of the sizes.
-zadapt :: forall value a b r. (ValueCons value, IsInteger a, IsInteger b, ShapeOf a ~ ShapeOf b)
-     => value a -> CodeGenFunction r (value b)
+zadapt ::
+    forall a b r.
+    (IsInteger a, IsInteger b, ShapeOf a ~ ShapeOf b) =>
+    Value a -> CodeGenFunction r (Value b)
 zadapt =
    case compare (sizeOf (typeDesc (LP.Proxy :: LP.Proxy a)))
                 (sizeOf (typeDesc (LP.Proxy :: LP.Proxy b))) of
-      LT -> convert FFI.constZExt FFI.buildZExt
-      EQ -> convert FFI.constBitCast FFI.buildBitCast
-      GT -> convert FFI.constTrunc FFI.buildTrunc
+      LT -> convertValue LP.Proxy FFI.buildZExt
+      EQ -> convertValue LP.Proxy FFI.buildBitCast
+      GT -> convertValue LP.Proxy FFI.buildTrunc
 
 -- | It is 'sext', 'trunc' or nop depending on the relation of the sizes.
-sadapt :: forall value a b r. (ValueCons value, IsInteger a, IsInteger b, ShapeOf a ~ ShapeOf b)
-     => value a -> CodeGenFunction r (value b)
+sadapt ::
+    forall a b r.
+    (IsInteger a, IsInteger b, ShapeOf a ~ ShapeOf b) =>
+    Value a -> CodeGenFunction r (Value b)
 sadapt =
    case compare (sizeOf (typeDesc (LP.Proxy :: LP.Proxy a)))
                 (sizeOf (typeDesc (LP.Proxy :: LP.Proxy b))) of
-      LT -> convert FFI.constSExt FFI.buildSExt
-      EQ -> convert FFI.constBitCast FFI.buildBitCast
-      GT -> convert FFI.constTrunc FFI.buildTrunc
+      LT -> convertValue LP.Proxy FFI.buildSExt
+      EQ -> convertValue LP.Proxy FFI.buildBitCast
+      GT -> convertValue LP.Proxy FFI.buildTrunc
 
 -- | It is 'sadapt' or 'zadapt' depending on the sign mode.
-adapt :: forall value a b r. (ValueCons value, IsInteger a, IsInteger b, ShapeOf a ~ ShapeOf b, Signed a ~ Signed b)
-     => value a -> CodeGenFunction r (value b)
+adapt ::
+    forall a b r.
+    (IsInteger a, IsInteger b, ShapeOf a ~ ShapeOf b, Signed a ~ Signed b) =>
+    Value a -> CodeGenFunction r (Value b)
 adapt =
    case compare (sizeOf (typeDesc (LP.Proxy :: LP.Proxy a)))
                 (sizeOf (typeDesc (LP.Proxy :: LP.Proxy b))) of
       LT ->
          if isSigned (LP.Proxy :: LP.Proxy b)
-           then convert FFI.constSExt FFI.buildSExt
-           else convert FFI.constZExt FFI.buildZExt
-      EQ -> convert FFI.constBitCast FFI.buildBitCast
-      GT -> convert FFI.constTrunc FFI.buildTrunc
+           then convertValue LP.Proxy FFI.buildSExt
+           else convertValue LP.Proxy FFI.buildZExt
+      EQ -> convertValue LP.Proxy FFI.buildBitCast
+      GT -> convertValue LP.Proxy FFI.buildTrunc
 
 -- | Truncate a floating point value.
-fptrunc :: (ValueCons value, IsFloating a, IsFloating b, ShapeOf a ~ ShapeOf b, IsSized a, IsSized b, SizeOf a :>: SizeOf b)
-        => value a -> CodeGenFunction r (value b)
-fptrunc = convert FFI.constFPTrunc FFI.buildFPTrunc
+fptrunc ::
+    (IsFloating a, IsFloating b, ShapeOf a ~ ShapeOf b,
+     IsSized a, IsSized b, SizeOf a :>: SizeOf b) =>
+    Value a -> CodeGenFunction r (Value b)
+fptrunc = convertValue LP.Proxy FFI.buildFPTrunc
 
 -- | Extend a floating point value.
-fpext :: (ValueCons value, IsFloating a, IsFloating b, ShapeOf a ~ ShapeOf b, IsSized a, IsSized b, SizeOf a :<: SizeOf b)
-      => value a -> CodeGenFunction r (value b)
-fpext = convert FFI.constFPExt FFI.buildFPExt
+fpext ::
+    (IsFloating a, IsFloating b, ShapeOf a ~ ShapeOf b,
+     IsSized a, IsSized b, SizeOf a :<: SizeOf b) =>
+    Value a -> CodeGenFunction r (Value b)
+fpext = convertValue LP.Proxy FFI.buildFPExt
 
 {-# DEPRECATED fptoui "use fptoint since it is type-safe with respect to signs" #-}
 -- | Convert a floating point value to an unsigned integer.
-fptoui :: (ValueCons value, IsFloating a, IsInteger b, ShapeOf a ~ ShapeOf b) => value a -> CodeGenFunction r (value b)
-fptoui = convert FFI.constFPToUI FFI.buildFPToUI
+fptoui ::
+    (IsFloating a, IsInteger b, ShapeOf a ~ ShapeOf b) =>
+    Value a -> CodeGenFunction r (Value b)
+fptoui = convertValue LP.Proxy FFI.buildFPToUI
 
 {-# DEPRECATED fptosi "use fptoint since it is type-safe with respect to signs" #-}
 -- | Convert a floating point value to a signed integer.
-fptosi :: (ValueCons value, IsFloating a, IsInteger b, ShapeOf a ~ ShapeOf b) => value a -> CodeGenFunction r (value b)
-fptosi = convert FFI.constFPToSI FFI.buildFPToSI
+fptosi ::
+    (IsFloating a, IsInteger b, ShapeOf a ~ ShapeOf b) =>
+    Value a -> CodeGenFunction r (Value b)
+fptosi = convertValue LP.Proxy FFI.buildFPToSI
 
 -- | Convert a floating point value to an integer.
 -- It is mapped to @fptosi@ or @fptoui@ depending on the type @a@.
-fptoint :: forall value a b r. (ValueCons value, IsFloating a, IsInteger b, ShapeOf a ~ ShapeOf b) => value a -> CodeGenFunction r (value b)
+fptoint ::
+    forall a b r.
+    (IsFloating a, IsInteger b, ShapeOf a ~ ShapeOf b) =>
+    Value a -> CodeGenFunction r (Value b)
 fptoint =
    if isSigned (LP.Proxy :: LP.Proxy b)
-     then convert FFI.constFPToSI FFI.buildFPToSI
-     else convert FFI.constFPToUI FFI.buildFPToUI
+     then convertValue LP.Proxy FFI.buildFPToSI
+     else convertValue LP.Proxy FFI.buildFPToUI
 
 
 {- DEPRECATED uitofp "use inttofp since it is type-safe with respect to signs" -}
 -- | Convert an unsigned integer to a floating point value.
 -- Although 'inttofp' should be prefered, this function may be useful for conversion from Bool.
-uitofp :: (ValueCons value, IsInteger a, IsFloating b, ShapeOf a ~ ShapeOf b) => value a -> CodeGenFunction r (value b)
-uitofp = convert FFI.constUIToFP FFI.buildUIToFP
+uitofp ::
+    (IsInteger a, IsFloating b, ShapeOf a ~ ShapeOf b) =>
+    Value a -> CodeGenFunction r (Value b)
+uitofp = convertValue LP.Proxy FFI.buildUIToFP
 
 {- DEPRECATED sitofp "use inttofp since it is type-safe with respect to signs" -}
 -- | Convert a signed integer to a floating point value.
 -- Although 'inttofp' should be prefered, this function may be useful for conversion from Bool.
-sitofp :: (ValueCons value, IsInteger a, IsFloating b, ShapeOf a ~ ShapeOf b) => value a -> CodeGenFunction r (value b)
-sitofp = convert FFI.constSIToFP FFI.buildSIToFP
+sitofp ::
+    (IsInteger a, IsFloating b, ShapeOf a ~ ShapeOf b) =>
+    Value a -> CodeGenFunction r (Value b)
+sitofp = convertValue LP.Proxy FFI.buildSIToFP
 
 -- | Convert an integer to a floating point value.
 -- It is mapped to @sitofp@ or @uitofp@ depending on the type @a@.
-inttofp :: forall value a b r. (ValueCons value, IsInteger a, IsFloating b, ShapeOf a ~ ShapeOf b) => value a -> CodeGenFunction r (value b)
+inttofp ::
+    forall a b r.
+    (IsInteger a, IsFloating b, ShapeOf a ~ ShapeOf b) =>
+    Value a -> CodeGenFunction r (Value b)
 inttofp =
    if isSigned (LP.Proxy :: LP.Proxy a)
-     then convert FFI.constSIToFP FFI.buildSIToFP
-     else convert FFI.constUIToFP FFI.buildUIToFP
+     then convertValue LP.Proxy FFI.buildSIToFP
+     else convertValue LP.Proxy FFI.buildUIToFP
 
 
 -- | Convert a pointer to an integer.
@@ -740,41 +775,37 @@ bitcast = convert FFI.constBitCast FFI.buildBitCast
 
 --------------------------------------
 
-type CmpValueResult value0 value1 a = BinOpValue value0 value1 (CmpResult a)
-
 type CmpResult c = ShapedType (ShapeOf c) Bool
 
 class (IsFirstClass c) => CmpRet c where
     cmpBld :: LP.Proxy c -> CmpPredicate -> FFIBinOp
-    cmpCnst :: LP.Proxy c -> CmpPredicate -> FFIConstBinOp
 
-instance CmpRet Float   where cmpBld _ = fcmpBld ; cmpCnst _ = fcmpCnst
-instance CmpRet Double  where cmpBld _ = fcmpBld ; cmpCnst _ = fcmpCnst
-instance CmpRet FP128   where cmpBld _ = fcmpBld ; cmpCnst _ = fcmpCnst
-instance CmpRet Bool    where cmpBld _ = ucmpBld ; cmpCnst _ = ucmpCnst
-instance CmpRet Word    where cmpBld _ = ucmpBld ; cmpCnst _ = ucmpCnst
-instance CmpRet Word8   where cmpBld _ = ucmpBld ; cmpCnst _ = ucmpCnst
-instance CmpRet Word16  where cmpBld _ = ucmpBld ; cmpCnst _ = ucmpCnst
-instance CmpRet Word32  where cmpBld _ = ucmpBld ; cmpCnst _ = ucmpCnst
-instance CmpRet Word64  where cmpBld _ = ucmpBld ; cmpCnst _ = ucmpCnst
-instance CmpRet Int     where cmpBld _ = scmpBld ; cmpCnst _ = scmpCnst
-instance CmpRet Int8    where cmpBld _ = scmpBld ; cmpCnst _ = scmpCnst
-instance CmpRet Int16   where cmpBld _ = scmpBld ; cmpCnst _ = scmpCnst
-instance CmpRet Int32   where cmpBld _ = scmpBld ; cmpCnst _ = scmpCnst
-instance CmpRet Int64   where cmpBld _ = scmpBld ; cmpCnst _ = scmpCnst
+instance CmpRet Float   where cmpBld _ = fcmpBld
+instance CmpRet Double  where cmpBld _ = fcmpBld
+instance CmpRet FP128   where cmpBld _ = fcmpBld
+instance CmpRet Bool    where cmpBld _ = ucmpBld
+instance CmpRet Word    where cmpBld _ = ucmpBld
+instance CmpRet Word8   where cmpBld _ = ucmpBld
+instance CmpRet Word16  where cmpBld _ = ucmpBld
+instance CmpRet Word32  where cmpBld _ = ucmpBld
+instance CmpRet Word64  where cmpBld _ = ucmpBld
+instance CmpRet Int     where cmpBld _ = scmpBld
+instance CmpRet Int8    where cmpBld _ = scmpBld
+instance CmpRet Int16   where cmpBld _ = scmpBld
+instance CmpRet Int32   where cmpBld _ = scmpBld
+instance CmpRet Int64   where cmpBld _ = scmpBld
 instance CmpRet (Foreign.Ptr a)
-                        where cmpBld _ = ucmpBld ; cmpCnst _ = ucmpCnst
+                        where cmpBld _ = ucmpBld
 instance (IsType a) =>
-         CmpRet (Ptr a) where cmpBld _ = ucmpBld ; cmpCnst _ = ucmpCnst
+         CmpRet (Ptr a) where cmpBld _ = ucmpBld
 
 instance (Dec.Positive n) => CmpRet (WordN n) where
-    cmpBld _ = ucmpBld ; cmpCnst _ = ucmpCnst
+    cmpBld _ = ucmpBld
 instance (Dec.Positive n) => CmpRet (IntN n) where
-    cmpBld _ = scmpBld ; cmpCnst _ = scmpCnst
+    cmpBld _ = scmpBld
 
 instance (CmpRet a, IsPrimitive a, Dec.Positive n) => CmpRet (Vector n a) where
     cmpBld _ = cmpBld (LP.Proxy :: LP.Proxy a)
-    cmpCnst _ = cmpCnst (LP.Proxy :: LP.Proxy a)
 
 
 {- |
@@ -785,14 +816,11 @@ that is @NaN@ operands yields 'False' as result.
 Pointers are compared unsigned.
 These choices are consistent with comparison in plain Haskell.
 -}
-cmp :: forall value0 value1 a r.
-   (ValueCons2 value0 value1, CmpRet a) =>
-   CmpPredicate -> value0 a -> value1 a ->
-   CodeGenFunction r (CmpValueResult value0 value1 a)
-cmp p =
-    binop
-        (cmpCnst (LP.Proxy :: LP.Proxy a) p)
-        (cmpBld (LP.Proxy :: LP.Proxy a) p)
+cmp :: forall a r.
+   (CmpRet a) =>
+   CmpPredicate -> Value a -> Value a ->
+   CodeGenFunction r (Value (CmpResult a))
+cmp p = binopValue (cmpBld (LP.Proxy :: LP.Proxy a) p)
 
 ucmpBld :: CmpPredicate -> FFIBinOp
 ucmpBld p = flip FFI.buildICmp (FFI.fromIntPredicate (uintFromCmpPredicate p))
@@ -804,58 +832,27 @@ fcmpBld :: CmpPredicate -> FFIBinOp
 fcmpBld p = flip FFI.buildFCmp (FFI.fromRealPredicate (fpFromCmpPredicate p))
 
 
-ucmpCnst :: CmpPredicate -> FFIConstBinOp
-ucmpCnst p = FFI.constICmp (FFI.fromIntPredicate (uintFromCmpPredicate p))
-
-scmpCnst :: CmpPredicate -> FFIConstBinOp
-scmpCnst p = FFI.constICmp (FFI.fromIntPredicate (sintFromCmpPredicate p))
-
-fcmpCnst :: CmpPredicate -> FFIConstBinOp
-fcmpCnst p = FFI.constFCmp (FFI.fromRealPredicate (fpFromCmpPredicate p))
-
-
-_ucmp ::
-    (ValueCons2 value0 value1, CmpRet a, IsInteger a) =>
-    CmpPredicate -> value0 a -> value1 a ->
-    CodeGenFunction r (CmpValueResult value0 value1 a)
-_ucmp p = binop (ucmpCnst p) (ucmpBld p)
-
-_scmp ::
-    (ValueCons2 value0 value1, CmpRet a, IsInteger a) =>
-    CmpPredicate -> value0 a -> value1 a ->
-    CodeGenFunction r (CmpValueResult value0 value1 a)
-_scmp p = binop (scmpCnst p) (scmpBld p)
-
 pcmp ::
-    (ValueCons2 value0 value1, IsType a) =>
-    IntPredicate -> value0 (Ptr a) -> value1 (Ptr a) ->
-    CodeGenFunction r (BinOpValue value0 value1 (Ptr a))
-pcmp p =
-    binop
-        (FFI.constICmp (FFI.fromIntPredicate p))
-        (flip FFI.buildICmp (FFI.fromIntPredicate p))
+    (IsType a) =>
+    IntPredicate -> Value (Ptr a) -> Value (Ptr a) ->
+    CodeGenFunction r (Value Bool)
+pcmp p = binopValue (flip FFI.buildICmp (FFI.fromIntPredicate p))
 
 
 {-# DEPRECATED icmp "use cmp or pcmp instead" #-}
 -- | Compare integers.
 icmp ::
-    (ValueCons2 value0 value1, CmpRet a, IsIntegerOrPointer a) =>
-    IntPredicate -> value0 a -> value1 a ->
-    CodeGenFunction r (CmpValueResult value0 value1 a)
-icmp p =
-    binop
-        (FFI.constICmp (FFI.fromIntPredicate p))
-        (flip FFI.buildICmp (FFI.fromIntPredicate p))
+    (CmpRet a, IsIntegerOrPointer a) =>
+    IntPredicate -> Value a -> Value a ->
+    CodeGenFunction r (Value (CmpResult a))
+icmp p = binopValue (flip FFI.buildICmp (FFI.fromIntPredicate p))
 
 -- | Compare floating point values.
 fcmp ::
-    (ValueCons2 value0 value1, CmpRet a, IsFloating a) =>
-    FPPredicate -> value0 a -> value1 a ->
-    CodeGenFunction r (CmpValueResult value0 value1 a)
-fcmp p =
-    binop
-        (FFI.constFCmp (FFI.fromRealPredicate p))
-        (flip FFI.buildFCmp (FFI.fromRealPredicate p))
+    (CmpRet a, IsFloating a) =>
+    FPPredicate -> Value a -> Value a ->
+    CodeGenFunction r (Value (CmpResult a))
+fcmp p = binopValue (flip FFI.buildFCmp (FFI.fromRealPredicate p))
 
 --------------------------------------
 
